""" Used to build and return Athena queries. """


def create_table(bucketname):
    """ Returns query for behold table in Athena. """
    query_string = f"""CREATE EXTERNAL TABLE behold (
            eventversion STRING,
            useridentity STRUCT<
                        type:STRING,
                        principalid:STRING,
                        arn:STRING,
                        accountid:STRING,
                        invokedby:STRING,
                        accesskeyid:STRING,
                        userName:STRING,
            sessioncontext:STRUCT<
            attributes:STRUCT<
                        mfaauthenticated:STRING,
                        creationdate:STRING>,
            sessionissuer:STRUCT<
                        type:STRING,
                        principalId:STRING,
                        arn:STRING,
                        accountId:STRING,
                        userName:STRING>>>,
            eventtime STRING,
            eventsource STRING,
            eventname STRING,
            awsregion STRING,
            sourceipaddress STRING,
            useragent STRING,
            errorcode STRING,
            errormessage STRING,
            requestparameters STRING,
            responseelements STRING,
            additionaleventdata STRING,
            requestid STRING,
            eventid STRING,
            resources ARRAY<STRUCT<
                        ARN:STRING,
                        accountId:STRING,
                        type:STRING>>,
            eventtype STRING,
            apiversion STRING,
            readonly STRING,
            recipientaccountid STRING,
            serviceeventdetails STRING,
            sharedeventid STRING,
            vpcendpointid STRING
            )
            PARTITIONED BY (account string, region string, year string, month string)
            ROW FORMAT SERDE 'com.amazon.emr.hive.serde.CloudTrailSerde'
            STORED AS INPUTFORMAT 'com.amazon.emr.cloudtrail.CloudTrailInputFormat'
            OUTPUTFORMAT 'org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat'
            LOCATION 's3://{bucketname}/AWSLogs/';
            """
    return (query_string, "setup/create_table")


def add_to_partition(cloudtrail_bucket, account, region, year, month):
    """ Returns query for adding partition to behold Athena table. """
    query_string = f"""ALTER TABLE behold
        ADD PARTITION (account='{account}', region='{region}', year='{year}', month='{month}')
        LOCATION 's3://{cloudtrail_bucket}/AWSLogs/{account}/CloudTrail/{region}/{year}/{month}/';"""
    return (query_string, f"setup/add_to_partition/{account}-{region}-{year}-{month}")


def active_roles(account, days_back):
    """ Returns query for finding active roles (since days_back value). """
    query_string = f"""SELECT DISTINCT useridentity.sessioncontext.sessionissuer.arn
        FROM behold
        WHERE account = '{account}'
        AND useridentity.type = 'AssumedRole'
        AND from_iso8601_timestamp(eventtime) > date_add('day', -{days_back}, now());"""
    return (query_string, f"athena_results/active_roles/{account}")


def active_users(account, days_back):
    """ Returns query for finding active users (since days_back value)."""
    query_string = f"""SELECT DISTINCT useridentity.arn
        FROM behold
        WHERE account = '{account}'
        AND useridentity.type = 'IAMUser'
        AND useridentity.arn IS NOT NULL
        AND from_iso8601_timestamp(eventtime) > date_add('day', -{days_back}, now());"""
    return (query_string, f"athena_results/active_users/{account}")


def services_by_role(account, days_back, role_arn, role_name):
    """ Returns query for eventsource (service) / actions performed by role. """
    query_string = f"""SELECT DISTINCT eventsource, eventname FROM behold
        WHERE account = '{account}'
        AND (useridentity.sessioncontext.sessionissuer.arn = '{role_arn}')
        AND from_iso8601_timestamp(eventtime) > date_add('day', -{days_back}, now())
        ORDER BY eventsource, eventname;"""
    return (query_string, f"athena_results/services_by_role/{account}/{role_name}")


def services_by_user(account, days_back, user_arn, user_name):
    """ Returns query for eventsource (service) / actions performed by user. """
    query_string = f"""SELECT DISTINCT eventsource, eventname FROM behold
        WHERE account = '{account}'
        AND (useridentity.arn = '{user_arn}')
        AND from_iso8601_timestamp(eventtime) > date_add('day', -{days_back}, now())
        ORDER BY eventsource, eventname;"""
    return (query_string, f"athena_results/services_by_user/{account}/{user_name}")



import boto3
import logging
from . import athena_query_strings
from . import s3
from . import utils

logger = logging.getLogger(__name__)


class Athena():
    """ Class for interacting with Athena service in AWS. """
    def __init__(self, metadata, session=None):
        """ Sets required variables.  Creates Athena boto3 client. """
        self.metadata = metadata
        self.accounts = metadata['accounts_to_partition']
        self.days_back = metadata['days_back']
        self.cloudtrail_bucket = metadata['cloudtrail_bucket']
        self.behold_bucket = metadata['behold_bucket']
        self.region = metadata['region']
        self.s3 = s3.S3(metadata, session)
        if session is None:
            self.client = boto3.client('athena', region_name=self.region)
        else:
            self.client = session.client('athena')

    def active_resources(self):
        """ Creates list objects which are used to store location to Athena output files.
        Executes Athena queries to determine which roles and users have been used since days_back. """
        self.active_roles_output_files = []
        self.active_users_output_files = []
        self.active_roles_query()
        self.active_users_query()

    def start_query_execution(self, query_string, path):
        """ Takes Athena query string and output path and executes the query. """
        response = self.client.start_query_execution(
            QueryString=query_string,
            ResultConfiguration={
                "OutputLocation": f"s3://{self.behold_bucket}/{path}"
            }
        )
        if response['ResponseMetadata']['HTTPStatusCode'] == 200:
            return response['QueryExecutionId']
        else:
            logger.error(f"Response failed:\n{response}")

    def set_up_table_and_partitions(self):
        """ Sets up partitions to be used in Athena table. """
        logger.info("Setting up Athena table.")
        query_string, path = athena_query_strings.create_table(
            self.cloudtrail_bucket)
        self.start_query_execution(query_string, path)
        years_months_list = utils.year_month_parser(days_back=self.days_back)
        for account in self.metadata['accounts_to_partition']:
            for region in self.metadata['regions_to_partition']:
                for year, month in years_months_list:
                    logger.info(f"Adding partition to Athena table: {account} | {region} | {year} | {month}")
                    query_string, path = athena_query_strings.add_to_partition(
                        cloudtrail_bucket=self.cloudtrail_bucket,
                        account=account,
                        region=region,
                        year=year,
                        month=month
                    )
                    self.start_query_execution(query_string, path)

    def active_roles_query(self):
        """ Runs query to determine which roles have been used since days_back.
        Stores the path to Athena output file in active_roles_output_files list. """
        for account in self.accounts:
            logger.info(f"Querying Athena for active roles in {account} (past {self.days_back} days).")
            query_string, path = athena_query_strings.active_roles(
                account=account,
                days_back=self.days_back
            )
            execution_id = self.start_query_execution(query_string, path)
            output_dict = {
                "account": account,
                "path": f"{path}/{execution_id}.csv"
            }
            self.active_roles_output_files.append(output_dict)
            self.s3.check_object_exists(
                bucket=self.behold_bucket,
                key=f"{path}/{execution_id}.csv"
            )

    def active_users_query(self):
        """ Runs query to determine which users have been used since days_back.
        Stores the path to Athena output file in active_users_output_files list. """
        for account in self.accounts:
            logger.info(f"Querying Athena for active users in {account} (past {self.days_back} days).")
            query_string, path = athena_query_strings.active_users(
                account=account,
                days_back=self.days_back
            )
            execution_id = self.start_query_execution(query_string, path)
            output_dict = {
                "account": account,
                "path": f"{path}/{execution_id}.csv"
            }
            self.active_users_output_files.append(output_dict)
            self.s3.check_object_exists(
                bucket=self.behold_bucket,
                key=f"{path}/{execution_id}.csv"
            )

    def services_by_role_query(self, account, list_of_arns):
        """ Runs query to determine which services / actions have been used by a role.
        Stores the path to Athena output file in services_by_role_output_files list. """
        logger.info(f"Querying Athena for services used by role in {account}.")
        self.services_by_role_output_files = []
        for role_arn in list_of_arns:
            role_name = role_arn.split('/')
            role_name = role_name[len(role_name) - 1]
            logger.info(f"Querying by role: {role_name}")
            query_string, path = athena_query_strings.services_by_role(
                account=account,
                days_back=self.days_back,
                role_arn=role_arn,
                role_name=role_name
            )
            execution_id = self.start_query_execution(query_string, path)
            output_dict = {
                "account": account,
                "role_arn": role_arn,
                "name": role_name,
                "path": f"{path}/{execution_id}.csv"
            }
            self.services_by_role_output_files.append(output_dict)
            self.s3.check_object_exists(
                bucket=self.behold_bucket,
                key=f"{path}/{execution_id}.csv"
            )

    def services_by_user_query(self, account, list_of_arns):
        """ Runs query to determine which services / actions have been used by a role.
        Stores the path to Athena output file in services_by_role_output_files list. """
        logger.info(f"Querying Athena for services used by user in {account}")
        self.services_by_user_output_files = []
        for user_arn in list_of_arns:
            user_name = user_arn.split('/')[1]
            logger.info(f"Querying by user: {user_name}")
            query_string, path = athena_query_strings.services_by_user(
                account=account,
                user_arn=user_arn,
                user_name=user_name,
                days_back=self.days_back
            )
            execution_id = self.start_query_execution(query_string, path)
            output_dict = {
                "account": account,
                "user_arn": user_arn,
                "name": user_name,
                "path": f"{path}/{execution_id}.csv"
            }
            self.services_by_user_output_files.append(output_dict)
            self.s3.check_object_exists(
                bucket=self.behold_bucket,
                key=f"{path}/{execution_id}.csv"
            )



import boto3
import datetime


def year_month_parser(days_back=30):
    """ Calculates a list of tuples formatted in (year, month) given days_back """
    year_month = set()
    for x in range(days_back):
        date = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=x)
        year_month_tuple = date.strftime("%Y"), date.strftime("%m")
        year_month.add(year_month_tuple)
    return list(year_month)


class Boto():
    def __init__(self, metadata):
        self.region = metadata['region']
        self.session = boto3.session.Session(region_name=self.region)

    def get_cloudtrail_regions(self):
        """ Uses boto3 session to return list of supported regions for CloudTrail. """
        cloudtrail_regions = self.session.get_available_regions('cloudtrail')
        return cloudtrail_regions

    def get_account(self):
        """ Uses sts get_caller_identity function to return local account ID. """
        self.stsclient = boto3.client('sts')
        account = self.stsclient.get_caller_identity()['Account']
        return account



import csv


def single_column_csv_to_list(string):
    """ Takes a CSV string from a CSV which has only one column.
    Converts this string into a list, omitting the header column. """
    reader = csv.reader(string.splitlines())
    list_output = []
    for row in reader:
        list_output.extend(row)
    list_output.pop(0)
    return list_output


def csv_to_list_of_dicts(string):
    """ Takes a CSV string.  Creates a list of dictionaries by
    using the headers of the CSV as keys, and the values in each row as values. """
    reader = csv.DictReader(string.splitlines())
    list_of_dicts = []
    for row in reader:
        list_of_dicts.append(dict(row))
    return list_of_dicts



import boto3
import botocore
import time
import logging

logger = logging.getLogger(__name__)


class S3():
    """ Class for interacting with S3 service in AWS. """
    def __init__(self, metadata, session=None):
        """ Sets region and creates boto3 client. """
        self.region = metadata['region']
        self.metadata = metadata
        if session is None:
            self.client = boto3.client('s3', region_name=self.region)
        else:
            self.client = session.client('s3')

    def check_object_exists(self, bucket, key):
        obj_exists = False
        while not obj_exists:
            try:
                self.client.head_object(
                    Bucket=bucket,
                    Key=key
                )
                obj_exists = True
            except botocore.exceptions.ClientError:
                time.sleep(.500)
        return True

    def get_object(self, bucket, key):
        """ Checks if object exists.  If it doesn't, wait half a second and check again.
        Once the object is found it is downloaded."""
        logger.info(f"Downloading {key} from {bucket}.")

        self.check_object_exists(bucket, key)

        obj = self.client.get_object(
            Bucket=bucket,
            Key=key
        )
        return obj['Body'].read().decode()

    def put_object(self, bucket, key, encoded_object):
        """ Uploads object to S3. """
        logger.info(f"Uploading {key} to {bucket}.")
        self.client.put_object(
            Body=encoded_object,
            Bucket=bucket,
            Key=key
        )



from . import utils
from datetime import datetime
import os
import logging
import sys
import yaml

logger = logging.getLogger(__name__)


def read(filename):
    """ Reads metadata file.  Returns dict after performing validation and setting defaults. """
    logger.info(f"Reading {filename} from metadata folder.")
    metadata = os.path.abspath(filename)

    with open(metadata, "r") as meta:
        metadata = yaml.load(meta.read())

    validate_metadata(metadata)

    return metadata


def validate_metadata(metadata):
    """ Ensuring required values are set. """
    if 'region' not in metadata:
        sys.exit("'region' must be included in metadata file.\n" +
                 "'region' specifies the region in which to create the Athena table.\n" +
                 "'region' should be in the same region as your CloudTrail bucket.")
    elif 'cloudtrail_bucket' not in metadata:
        sys.exit("'cloudtrail_bucket' must be included in metadata file.\n" +
                 "'cloudtrail_bucket' specifies the name of your CloudTrail bucket.")
    elif 'behold_bucket' not in metadata:
        sys.exit("'behold_bucket' must be included in metadata file.\n" +
                 "'behold_bucket' specifies the bucket that behold will store its output files.\n" +
                 "'behold_bucket' should be in the same region as your CloudTrail bucket.")


def set_defaults(metadata, boto=None):
    """ Sets defaults for non-required values if they are not specified in the metadata file. """
    if boto is None:
        boto = utils.Boto(metadata)
    if 'years_to_partition' not in metadata:
        current_year = datetime.now().year
        logger.info(f"Setting 'years_to_partition' to {current_year}")
        metadata['years_to_partition'] = [current_year]
    if 'days_back' not in metadata:
        logger.info(f"Setting 'days_back' to 30.")
        metadata['days_back'] = 30
    if 'regions_to_partition' not in metadata:
        logger.info("Setting 'regions_to_partition' to all supported CloudTrail regions.")
        metadata['regions_to_partition'] = boto.get_cloudtrail_regions()
    if 'accounts_to_partition' not in metadata:
        local_account = boto.get_account()
        logger.info(f"Setting 'accounts_to_partition' to {local_account}")
        metadata['accounts_to_partition'] = [local_account]
    return metadata



import json
import logging
import requests

logger = logging.getLogger(__name__)


class PolicyGenerator():
    """ Class for parsing actions used by service. """
    def __init__(self):
        self.actions = []
        self.get_service_actions()

    def get_service_actions(self):
        """ Gets a list of possible actions from AWS Policy Generator site. """
        services = requests.get('https://awspolicygen.s3.amazonaws.com/js/policies.js').text.strip('app.PolicyEditorConfig=')
        services = json.loads(services)['serviceMap']
        for service in services:
            for action in services[service]['Actions']:
                self.actions.append(f"{services[service]['StringPrefix']}:{action}")

    def generate_list_of_actions(self, list_of_dicts):
        """ Generates a list of service:action to be used for creating IAM policies.
        CloudTrail records eventsource in format s3.amazonaws.com (using S3 as an example).
        Service is extracted from the full event source. """
        supported_actions = []
        unsupported_actions = []
        for dictionary in list_of_dicts:
            service = dictionary['eventsource'].split('.')[0]
            action = f"{service}:{dictionary['eventname']}"
            if action in self.actions:
                supported_actions.append(action)
            else:
                logger.info(f"{action} not supported.")
                unsupported_actions.append(action)

        supported_actions.sort()
        unsupported_actions.sort()
        return supported_actions, unsupported_actions

    @staticmethod
    def format_actions(list_of_actions):
        """ Returns formatted list of actions that will be outputted in .txt files. """
        formatted_actions = []
        for action in list_of_actions:
            formatted_actions.append(f'"{action}"')
        formatted_actions = ",\n".join(formatted_actions)
        return formatted_actions

    @staticmethod
    def build_policy(list_of_actions):
        """ Returns IAM policy using list of actions.  Will be outputted to .json files. """
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "behold",
                    "Effect": "Allow",
                    "Action": list_of_actions,
                    "Resource": "*"
                }
            ]
        }
        return json.dumps(policy)



from setuptools import setup, find_packages

setup(name='beholdiam',
      version='1.0.4',
      description='Behold uses Athena and CloudTrail to audit your IAM resources and generate policy recommendations.',
      url='https://github.com/kwkeefer/behold',
      author='Kyle Keefer',
      author_email='kyle.keefer@protonmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'boto3',
      ],
      entry_points={
          'console_scripts': [
              'behold=behold.behold:main'
          ]
      },
      zip_safe=False)



from .libs import metadata
from .libs import utils
from .libs.athena import Athena
from .libs.s3 import S3
from .libs.csv_parser import single_column_csv_to_list, csv_to_list_of_dicts
from .libs.policy_generator import PolicyGenerator
import argparse
import logging


def arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("metadata")
    parser.add_argument("--setup", action="store_true")
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()
    return args


def initialize_classes(args):
    """ Reading metadata, performing metadata validation, initializing required classes.
    Classes / metadata stored in initc dictionary. """
    initc = {}
    meta = metadata.read(args.metadata)
    initc['boto'] = utils.Boto(meta)
    initc['meta'] = metadata.set_defaults(meta, initc['boto'])
    initc['s3'] = S3(initc['meta'], initc['boto'].session)
    initc['athena'] = Athena(initc['meta'], initc['boto'].session)
    initc['policygen'] = PolicyGenerator()
    return initc


def get_arns_from_athena_output(users_or_roles, initc):
    """ Function to get list of arns of active users or roles. """
    if users_or_roles == "users":
        athena_output_files = initc['athena'].active_users_output_files
        services_by_query = initc['athena'].services_by_user_query
    elif users_or_roles == "roles":
        athena_output_files = initc['athena'].active_roles_output_files
        services_by_query = initc['athena'].services_by_role_query

    for dictionary in athena_output_files:
        obj = initc['s3'].get_object(initc['meta']["behold_bucket"], dictionary["path"])
        list_of_arns = single_column_csv_to_list(obj)
        initc['s3'].put_object(
            bucket=initc['meta']['behold_bucket'],
            key=f"behold_results/{dictionary['account']}/{users_or_roles}/active_{users_or_roles}_in"
                f"_last_{initc['meta']['days_back']}_days.txt",
            encoded_object="\n".join(list_of_arns).encode()
        )
        services_by_query(
            account=dictionary["account"],
            list_of_arns=list_of_arns
        )


def build_behold_output_files(users_or_roles, initc):
    """ Builds list of services/actions and IAM policy for each role or user. """
    if users_or_roles == "users":
        athena_services_by_output_files = initc['athena'].services_by_user_output_files
    elif users_or_roles == "roles":
        athena_services_by_output_files = initc['athena'].services_by_role_output_files

    for dictionary in athena_services_by_output_files:
        obj = initc['s3'].get_object(initc['meta']["behold_bucket"], dictionary["path"])
        list_of_dicts = csv_to_list_of_dicts(obj)
        path_to_output = f"behold_results/{dictionary['account']}/{users_or_roles}/{dictionary['name']}/{dictionary['name']}_"
        supported_actions, unsupported_actions = initc['policygen'].generate_list_of_actions(list_of_dicts)
        formatted_supported_actions = initc['policygen'].format_actions(supported_actions)
        initc['s3'].put_object(
            bucket=initc['meta']["behold_bucket"],
            key=path_to_output + "iam_actions.txt",
            encoded_object=formatted_supported_actions.encode()
        )
        policy = initc['policygen'].build_policy(supported_actions)
        initc['s3'].put_object(
            bucket=initc['meta']['behold_bucket'],
            key=path_to_output + "iam_policy.json",
            encoded_object=policy.encode()
        )
        if unsupported_actions:
            initc['s3'].put_object(
                bucket=initc['meta']['behold_bucket'],
                key=path_to_output + "unsupported_actions.txt",
                encoded_object="\n".join(unsupported_actions).encode()
            )


def main():
    args = arguments()
    if args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    logging.basicConfig(
        level=log_level,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
    )

    initc = initialize_classes(args)

    # If --setup flag is passed, the Athena table and partition tables are set up.
    # Only needs to be done once unless metadata is updated to add more accounts, regions, or years.
    if args.setup:
        initc['athena'].set_up_table_and_partitions()

    initc['athena'].active_resources()

    get_arns_from_athena_output("users", initc)
    get_arns_from_athena_output("roles", initc)
    build_behold_output_files("users", initc)
    build_behold_output_files("roles", initc)


if __name__ == '__main__':
    main()
